module audio;

import SDL;
import SDL_mixer;

import std.string;

class Error{}

class Audio{
  public:
	this(){
		if(Mix_OpenAudio(22050, AUDIO_S16SYS, 2, 4096) != 0)
			active = false; //throw new Error;
		else
			active = true;
	}

	~this(){
		if(active){
			if(!(mus is null)){
				Mix_HaltMusic();
				Mix_FreeMusic(mus);
			}

			for(int a = 0; a < 8; a++)
				if(sfx[a] != null)
					Mix_FreeChunk(sfx[a]);

			Mix_HaltChannel(-1);
			Mix_CloseAudio();
		}
	}

	void playEffect(char[] filename, bool loop = false){
		if(!active)
			return;

		loadEffect(filename);
		playEffect(loop);
	}

	void playEffect(int number, bool loop = false){
		if(!active)
			return;

		if(number >= 8 || number < 0)
			number = 0;

		sfxChannel = Mix_PlayChannel(-1, sfx[number], loop == true ? -1 : 0);
	}

	void loadEffect(char[] filename, int number = 0){
		if(!active)
			return;

		if(number >= 8 || number < 0)
			number = 0;

		if(!(sfx[number] is null)){
			stopEffect();
			Mix_FreeChunk(sfx[number]);
		}

		sfx[number] = Mix_LoadWAV(std.string.toStringz(filename));
	}

	void stopEffect(){
		if(!active)
			return;

		Mix_HaltChannel(sfxChannel);
	}

	void playMusic(char[] filename, bool loop = true){
		if(!active)
			return;

		loadMusic(filename);
		playMusic(loop);
	}

	void playMusic(bool loop = true){
		if(!active)
			return;

		if(Mix_PlayMusic(mus, loop == true ? -1 : 0) == -1)
			throw new Error;
		//	musicIsPlaying = false;
		else
			musicIsPlaying = true;
	}

	void loadMusic(char[] filename){
		if(!active)
			return;

		if(mus != null){
			stopMusic();
			Mix_FreeMusic(mus);
		}

		mus = Mix_LoadMUS(std.string.toStringz(filename));
	}

	void pauseMusic(){
		if(!active)
			return;

		if(musicIsPlaying){
			Mix_PauseMusic();
			musicIsPaused = true;
		}
	}

	void unpauseMusic(){
		if(!active)
			return;

		if(musicIsPaused){
			Mix_ResumeMusic();
			musicIsPaused = false;
		}
	}

	void stopMusic(){
		if(!active)
			return;

		Mix_HaltMusic();
	}


	void stopAll(){
		if(!active)
			return;

		Mix_HaltMusic();
		Mix_HaltChannel(-1);
	}

  private:
  	int sfxChannel;
	bool active;

	bool musicIsPaused;
	bool musicIsPlaying;

	Mix_Chunk* sfx[8];
	Mix_Music* mus;
}

	int Mix_PlayChannel(int channel, Mix_Chunk* chunk, int loops) {
		return Mix_PlayChannelTimed(channel,chunk,loops,-1);
	}
	Mix_Chunk * Mix_LoadWAV(char *file) {
		return Mix_LoadWAV_RW(SDL_RWFromFile(file, "rb"), 1);
	}
