//import dc;
import std.stdio;
import std.string;

//Console console;
bool quit;

class Lispy{
  public:
  	this(){
		Atom* nil = new Atom;
		nil.type = Atom.symbol;
		nil.constant = true;
		nil.symbolName = "NIL";
		internSymbol("NIL", nil);

		Atom* t = new Atom;
		t.type = Atom.symbol;
		t.constant = true;
		t.symbolName = "T";
		internSymbol("T", t);


		// register the default built-in functions
		registerFunction("PRINT", 1, &Lprint);
		registerFunction("SET", 2, &set);
	}

	// Built in function implementations

	Atom* Lprint(List* args){
		if(args.data !is null)
			writefln("%s", print(args.data));
		return symbols["NIL"];
	}

	Atom* set(List* args){
		if(args.data.type != Atom.symbol)
			throw new Exception("Can only set a symbol");
		return internSymbol(args.data.symbolName, args.next.data);
	}
	
//***********************************
// Lispy proper begins here

	Atom* internSymbol(char[] name, Atom* value){
		Atom** v = name in symbols;
		if(v)
			if((*v).constant)
				throw new Exception("Cannot write to a constant");
		symbols[name] = value;
		return value;
	}
	
	struct Atom{
		int type;
		bool constant;
		enum {
			list,
			symbol, func,
			string, integer, floatingPoint, ratio	// self evaluating
		}
		union {
			List* listData;
			char[] symbolName;
			char[] stringData;
			Function* functionData;
			int intData;
			float floatData;
			struct ratioData {
				int numerator;
				int denominator;
			}
		}
	}

	struct List{
		Atom* data;
		List* next;

		void append(Atom* what){
			if(data is null){
				data = what;
				return;
			}
			List* end = this;
			while(end.next !is null)
				end = end.next;

			end.next = new List;
			end.next.data = what;
		}

		int opApply(int delegate(inout List) dg){
			int result = 0;
			List* l;
			for(l = this; l != null; l = l.next){
				result = dg(*l);
				if(result)
					break;
			}
			return result;
		}

	}

	void ignoreWhitespace(char[] expression, inout int location){
		while(expression[location] == ' '
		   || expression[location] == '\t'
		   || expression[location] == '\n'
		     )
			location++;
	}

	char[] extractWord(char[] expression, inout int location){
		const char[] endingSymbols = "), \t\n";

//		ignoreWhitespace(expression, location);

		int start = location;
		bool quoting;

		outer: while(location < expression.length){
			char a = expression[location];
			if(a == '\"')
				quoting = !quoting;
			if(!quoting)
				foreach(b; endingSymbols)
					if(a == b)
						break outer;
			location++;
		}

		return expression[start..location];
	}


	Atom* read(char[] expression, inout int location){

		Atom* result;

		// Ignore opening whitespace here, since it is irrelevant
		ignoreWhitespace(expression, location);

		switch(expression[location]){
			case '\'':	// quote
				result = new Atom;
				result.type = Atom.list;
				result.listData = new List;
				result.listData.data = new Atom;
				result.listData.data.type = Atom.symbol;
				result.listData.data.symbolName = "QUOTE";

				location++;
				result.listData.append(read(expression, location));
			break;
			case '(':	// list
				result = new Atom;
				result.type = Atom.list;
				result.listData = new List;


				location++;
				if(expression[location] == ')') // the empty list (NIL)
					location++;
				else
				while(expression[location] != ')')
					result.listData.append(read(expression, location));
			break;
			case '\"':	// string
				char[] word = extractWord(expression, location);
				result = new Atom;
				result.type = Atom.string;
				result.stringData = word[1..(length-1)]; // leaving off the quotes
			break;
//			case '+':
//			case '-':
//			if(expression[location + 1] ==
//			break;
			default:	// some other primitive
				char[] word = extractWord(expression, location);
				result = new Atom;
				result.type = Atom.symbol;
				result.symbolName = toupper(word); //FIXME should be uppercase except when escaped
		}

		return result;
	}

	Atom* evaluate(Atom* expression, Atom*[char[]]* specialSymbols = null){
		switch(expression.type){
			case Atom.list:
				if(expression.listData.data is null) // the empty list evaluates to itself
					return expression;
				if(expression.listData.data.type != Atom.symbol)
					throw new Exception("Must use symbols for function call");
				switch(expression.listData.data.symbolName){ // Handle special symbols
					case "QUOTE":
						Atom* a = expression.listData.next.data;
						return a;
					break;
					case "FUNCTION":
						Function* f = expression.listData.next.data.symbolName in functions;
						if(f is null)
							throw new Exception("No such function");
						Atom* a = new Atom;
						a.type = Atom.func;
						a.functionData = f;
						return a;
					break;
					default:
						return callFunction(expression.listData.data, expression.listData.next);
				}
			break;
			case Atom.symbol:
				if(expression.symbolName[0] == ':')
					return expression;
				if(specialSymbols !is null){
					Atom** b = expression.symbolName in *specialSymbols;
					if(b)
						return *b;
				}
				Atom** a = expression.symbolName in symbols;
				if(a)
					return *a;
				return symbols["NIL"];
			break;
			case Atom.string:
				return expression;
			break;
		}
		return null;
	}

	char[] print(Atom* expression){
		char[] result;
		if(expression is null)
			return "NULL EXPRESSION";
		switch(expression.type){
			case Atom.symbol:
				result ~= expression.symbolName;
			break;
			case Atom.list:
				result ~= "(";
				foreach(b; *expression.listData){
					if(b.data)
						result ~= print(b.data);
					if(b.next)
						result ~= " ";
				}
				result ~= ")";
			break;
			case Atom.string:
				result ~= "\"";
				result ~= expression.stringData;;
				result ~= "\"";
			break;
		}
		return result;
	}


	struct Function{
		char[] name;
		List* arguments;
		int argumentsLength;

		enum {lispy, native}
		int type;

		union {
			List* code;
			Atom* delegate(List* args) nativeCode;
		}

		char[] comment;
	}


	void registerFunction(char[] name, List* args, List* code, char[] comment = ""){
		Function f;
		f.name = name;
		f.arguments = args;
		f.type = Function.lispy;
		f.code = code;
		f.comment = comment;

		functions[name] = f;
	}

	/*
	void registerFunction(char[] name, List* args, Atom* delegate(List* args) code, char[] comment = ""){
		Function f;
		f.name = name;
		f.arguments = args;
		f.type = Function.native;
		f.nativeCode = code;
		f.comment = comment;

		functions[name] = f;
	}
	*/

	void registerFunction(char[] name, int argsLength, Atom* delegate(List* args) code, char[] comment = ""){
	/*
		List* a = new List;
		List* c = a;
		List* p;
		foreach(arg; args){
			c.data = new Atom;
			c.data.type = Atom.symbol;
			c.data.symbolName = arg;

			p = c;
			c.next = new List;
			c = c.next;
		}

		p.next = null;
		registerFunction(name, a, code, comment);
	*/
		Function f;
		f.name = name;
		f.argumentsLength = argsLength;
		f.type = Function.native;
		f.nativeCode = code;
		f.comment = comment;

		functions[name] = f;

	}



	Atom* callFunction(Atom* name, List* args){
		if(name.type != Atom.symbol)
			throw new Exception("Not a function. Try a symbol instead.");

		Function* f = name.symbolName in functions;
		if(f is null)
			throw new Exception(name.symbolName ~ ": no such function");

		List* evalArgs = new List;
		if(args !is null){
			List* es = evalArgs;
			List* p = es;
			foreach(i; *args){
				if(i.data is null)
					break;
				es.data = evaluate(i.data);
				es.next = new List;
				p = es;
				es = es.next;
			}
			p.next = null;
		}


		if((*f).type == Function.lispy){
			Atom*[char[]] localSymbols;
			List* c = evalArgs;
			foreach(i;*((*f).arguments)){
				if(i.data is null)
					break;
				if(c is null)
					throw new Exception("Too few arguments given to function");
				if(i.data.type == Atom.symbol)
					localSymbols[i.data.symbolName] = c.data;
				c = c.next;
			}

			Atom* co = new Atom;
			co.type = Atom.list;
			co.listData = (*f).code;
			return evaluate(co, &localSymbols);
		}else{
			int l;
			for(List* n = args; n !is null; n = n.next)
				l++;
			if(l != (*f).argumentsLength)
				throw new Exception("Wrong number of arguments given to function");
			return (*f).nativeCode(evalArgs);
		}
	}

  protected:

  private:
	Atom*[char[]] symbols;
	Function[char[]] functions;

}

/*
int main(char[][] args){
	Lispy l = new Lispy;
	int location;
	Lispy.Atom* a = l.read(args[1], location);
	location = 0;
	Lispy.Atom* b = l.read(args[2], location);
//	writefln("%s", l.print(a));
	writefln("%s", l.print(l.evaluate(a)));
	writefln("%s", l.print(l.evaluate(b)));
	return 0;
}
*/

