import engine;

import std.file;
import std.stdio;
import std.math;

import tile;

class Map{
  public:
	this(char[] f){
		ubyte[] d = cast(ubyte[]) read(f);
		int offset;
		tiles = loadTileSet(d, offset);
		width = d[offset++] + 1;
		height =d[offset++] + 1;

		data = d[offset..length];
	}

	void draw(Painter p, int x, int y, int xoff = 0, int yoff = 0){
		p.fill(RGB(255,0,0));
		Point l = XY(-abs(xoff),-abs(yoff));

		int offset = y*width + x;

		for(int b = 0;b < (512/16 + (yoff != 0) * 2); b++){
			for(int a = 0; a < (512/16 + (xoff != 0) * 2); a++){
				int mx = x+a;
				int my = y+b;
				if(mx < 0)
					mx += width;
				if(mx >= width)
					mx -= width;
				if(my < 0)
					my += height;
				if(my >= width)
					my -= height;

				offset = my*width + mx;

				p.drawImage(l, tiles[data[offset]].frames[0]);
				l.x+=16;
				//offset++;
			}
//			offset = (y+b)*width + x;
			l.x=-xoff;
			l.y+=16;
		}
	}

	Tile[] loadTileSet(ubyte[] data, out int off){
		Tile[] temp;
		temp.length = data[0];
		int offset = 1;
		for(int a=0;a< data[0];a++){
			Point point = XY(0,0);
			temp[a].frames.length = 1;
			temp[a].frames[0] = new Image(16,16);
			scope Painter p = new Painter(temp[a].frames[0]);
			for(int b=0;b<16*16;b++){
				p.putpixel(point, RGB(data[offset+0], data[offset+1],data[offset+2]));
				offset+=3;
				point.x++;
				if(point.x == 16){
					point.x = 0;
					point.y++;
				}
			}
		temp[a].frames[0].toGL();	// cache the texture in vid mem now
		}

		off = offset;
		return temp;
	}

	int width(){
		return w;
	}

	int height(){
		return h;
	}

  private:
	void width(int wid){
		w = wid;
	}

	void height(int hei){
		h = hei;
	}

	int w;
	int h;
	ubyte[] data;
	Tile[] tiles;

	char[] bgm;

	ubyte flags;
	const ubyte LOOPS		= 0x1;
	const ubyte RANDOM_BATTLES	= 0x2;
	const ubyte CAN_SAVE		= 0x4;
	const ubyte NIGHT_ADVANCES	= 0x8;
}

