module screen;

import SDL;
import SDL_image;
import SDL_ttf;
import std.string;

import std.stdio;

extern(C){
	void glGetIntegerv(int, void*);
	void glMatrixMode(int);
	void glPushMatrix();
	void glLoadIdentity();
	void glOrtho(double, double, double, double, double, double);
	void glPopMatrix();
	void glEnable(int);
	void glDisable(int);
	void glClear(int);
	void glBegin(int);
	void glVertex2f(float, float);
	void glEnd();
	void glColor3b(ubyte, ubyte, ubyte);
	void glColor3i(int, int, int);
	void glColor3f(float, float, float);
	void glColor4f(float, float, float, float);

	uint glGetError();

	void glDeleteTextures(int, uint*);

	char* gluErrorString(uint);

	void glRasterPos2i(int, int);
	void glDrawPixels(int, int, uint, uint, void*);
	void glClearColor(float, float, float, float);



	void glGenTextures(uint, uint*);
	void glBindTexture(int, int);
	void glTexParameteri(uint, uint, int);
	void glTexImage2D(int, int, int, int, int, int, int, int, void*);


	void glTexCoord2f(float, float);
	void glVertex2i(int, int);
	void glBlendFunc (int, int);
	void glViewport(int, int, int, int);


	const uint GL_BLEND = 0x0be2;
	const uint GL_SRC_ALPHA = 0x0302;
	const uint GL_ONE_MINUS_SRC_ALPHA = 0x0303;


	const uint GL_UNSIGNED_BYTE = 0x1401;
	const uint GL_RGB = 0x1907;
	const uint GL_RGBA = 0x1908;
	const uint GL_TEXTURE_2D =   0x0DE1;
	const uint GL_TEXTURE_MIN_FILTER = 0x2801;
	const uint GL_NEAREST = 0x2600;
	const uint GL_LINEAR = 0x2601;
	const uint GL_TEXTURE_MAG_FILTER = 0x2800;

	const uint GL_NO_ERROR = 0;



	const int GL_VIEWPORT = 0x0BA2;
	const int GL_MODELVIEW = 0x1700;
	const int GL_TEXTURE = 0x1702;
	const int GL_PROJECTION = 0x1701;
	const int GL_DEPTH_TEST = 0x0B71;

	const int GL_COLOR_BUFFER_BIT = 0x00004000;
	const int GL_ACCUM_BUFFER_BIT = 0x00000200;
	const int GL_DEPTH_BUFFER_BIT = 0x00000100;

	const int GL_POINTS = 0x0000;
	const int GL_LINES =  0x0001;
	const int GL_LINE_LOOP = 0x0002;
	const int GL_LINE_STRIP = 0x0003;
	const int GL_TRIANGLES = 0x0004;
	const int GL_TRIANGLE_STRIP = 5;
	const int GL_TRIANGLE_FAN = 6;
	const int GL_QUADS = 7;
	const int GL_QUAD_STRIP = 8;
	const int GL_POLYGON = 9;

}

class Error{}

public struct Point{
	int x;
	int y;
}

Point XY(int x, int y){
	Point p;
	p.x = x;
	p.y = y;
	return p;
}

Point XY(float x, float y){
	Point p;
	p.x = cast(int)x;
	p.y = cast(int)y;
	return p;
}

public struct Color{
	int r;
	int g;
	int b;
	int a;
}

Color RGB(int r, int g, int b){
	Color c;
	c.r = r;
	c.g = g;
	c.b = b;
	c.a = 255;
	return c;
}

Color RGBA(int r, int g, int b, int a){
	Color c;
	c.r = r;
	c.g = g;
	c.b = b;
	c.a = a;
	return c;
}

class FontEngine{
  public:

	static const FontEngine instance;

	~this(){
		foreach(a; fonts)
			if(a != null)
				TTF_CloseFont(a);
		TTF_Quit();
	}

	void loadFont(char[] font, int size = 12, int index = 0){
		if(fonts[index] != null)
			freeFont(index);
		TTF_Font* temp;
		temp = TTF_OpenFont(std.string.toStringz(font), size);
		if(temp == null)
			throw new Error;

		fonts[index] = temp;
	}

	void freeFont(int index = 0){
		if(fonts[index] != null){
			TTF_CloseFont(fonts[index]);
			fonts[index] = null;
		}
	}

	Image renderText(char[] text, Color foreground = RGB(255,255,255), Color background = RGB(0,0,0), int font = 0){
		Image* a = text in cache;
		if(a !is null)
			return *a;
		SDL_Color f;
		f.r = foreground.r;
		f.g = foreground.g;
		f.b = foreground.b;
		f.unused = foreground.a;
		SDL_Color b;
		b.r = background.r;
		b.g = background.g;
		b.b = background.b;
		b.unused = background.a;
//		SDL_Surface* s = TTF_RenderText_Shaded(fonts[font], std.string.toStringz(text), f, b);
		SDL_Surface* s = TTF_RenderText_Blended(fonts[font], std.string.toStringz(text), f);
		Image i = new Image(s);
		cache[text] = i;

		return i;
	}

  private:
	static this() {
		instance = new FontEngine;
	}

	this(){
		if(TTF_Init() == -1)
			throw new Error;

	}

	TTF_Font* fonts[8];
	Image[char[]] cache;
}

interface Drawable{
  public:
	void flip();
	int width();
	int height();
	int bpp();
	/*
	uint toGL();
	float texWidth();
	float texHeight();
	*/
  protected:
	SDL_Surface* surface();
}

class Image : Drawable{
  public:
  	this(SDL_Surface* s){
		if(s == null)
			throw new Error;
		sur = s;
	}

	/// Loads an image with the filename checking to see if it has already been loaded into the cache
	/// loads it as read-only
	static Image load(char[] filename){

	}

	this(char[] filename){
		sur = IMG_Load(std.string.toStringz(filename));
		if(sur == null)
			throw new Error;
	}

	this(int wid, int hei){
		sur = SDL_CreateRGBSurface(SDL_SWSURFACE, wid, hei, 32, 0xff0000, 0x00ff00, 0x0000ff, 0xff000000);
		if(sur == null)
			throw new Error;
		t = false;
	}

	~this(){
		if(t)
			glDeleteTextures(1, &tex);
		SDL_FreeSurface(sur);
	}

	void flip(){

	}

	int width(){
		return surface.w;
	}

	int height(){
		return surface.h;
	}

	int bpp(){
		return 32;
	}

	uint toGL(){
		if(t)
			return tex;
		else{
			float f[4];
			tex = SDL_GL_LoadTexture(surface, f.ptr);
			t = true;
			texWidth = f[2];
			texHeight = f[3];
			return tex;
		}
	}
  protected:
	SDL_Surface* surface(){
		return sur;
	}
  private:
  	SDL_Surface* sur;
	uint tex;
	bool t;
	float texWidth;
	float texHeight;
}

bool useGL;

class Screen : Drawable{
  public:
	this(int xres = 1024, int yres = 768, int bpp = 24, bool oGL = false){//true){
	oGL = false;
	oGL = true;
		if(!oGL)
			screen = SDL_SetVideoMode(xres, yres, bpp/*32*/, SDL_SWSURFACE);
		else{
			SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
			SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
			SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
			SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
			SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
			screen = SDL_SetVideoMode(xres, yres, 0, SDL_OPENGL);
			//screen = SDL_SetVideoMode(xres, yres, bpp, SDL_OPENGL);

   			glMatrixMode(GL_PROJECTION);
   			glLoadIdentity();
   			//glOrtho(0, 1000, yres, 0, 0, 1);
   			glOrtho(0, xres, yres, 0, 0, 1);
   			glMatrixMode(GL_MODELVIEW);

			glDisable(GL_DEPTH_TEST);

			glEnable(GL_TEXTURE_2D);


			glEnable (GL_BLEND); glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


			glClearColor(0,0,0,0);

//			glViewport(0,0,1024,1024);
		}

		useGL = oGL;

		if(screen == null)
			throw new Error;

//		SDL_SetAlpha(screen, SDL_SRCALPHA | SDL_RLEACCEL, 128);

		xr = xres;
		yr = yres;
	}

	~this(){
	}

	void flip(){
		if(useGL)
			SDL_GL_SwapBuffers();
		else
			SDL_Flip(screen);
	}

	int width(){
		return xr;
	}

	int height(){
		return yr;
	}

	int bpp(){
		return 124;
	}
	/*
	uint toGL(){
		throw new Error;
	}
	float texWidth(){
		return 1.0;
	}
	float texHeight(){
		return 1.0;
	}
*/
  protected:
	SDL_Surface* surface(){
		return screen;
	}

  private:
	SDL_Surface* screen;
	int xr;
	int yr;
}

scope class Painter{
  public:
	this(Drawable d){
	/+
		in {
			assert(!(s is null));
		}
	+/
		s = d;

		if(s.bpp == 124 && useGL){
			return;
		}
		if ( SDL_MUSTLOCK(s.surface) ) {
			if ( SDL_LockSurface(s.surface) < 0 ) {
				throw new Error;
			}
			locked = true;
		}
	}

	~this(){
		if (locked){
			SDL_UnlockSurface(s.surface);
		}
		s.flip();
	}

	void putpixel(Point where, Color color){
//		if(color.a == 255)
//			return;
		int x = where.x;
		int y = where.y;
		if(x < 0 || x >= s.width || y < 0 || y >= s.height)
			return;



		if(useGL && s.bpp == 124){
		//	y = 480 - y;
			glBegin(GL_POINTS);
			glColor4f(cast(float)color.r/255.0, cast(float)color.g/255.0, cast(float)color.b/255.0, cast(float)color.a / 255.0);
			glVertex2f(cast(float)x, cast(float)y);
			glEnd();
			return;
		}


		ubyte *bufp;

		if(s.bpp == 32){
			bufp = cast(ubyte*)(cast(ubyte*)s.surface.pixels + y*s.surface.pitch + x * 4);
				bufp[3] = color.a;
			        bufp[2] = color.r;
			        bufp[1] = color.g;
        			bufp[0] = color.b;
		}else{
			bufp = cast(ubyte*)(cast(ubyte*)s.surface.pixels + y*s.surface.pitch + x * 3);
			if(color.a == 255){
			        bufp[2] = color.r;
		        	bufp[1] = color.g;
	        		bufp[0] = color.b;
			}
			else{
			        bufp[2] = bufp[2] * (255-color.a) + (color.r * (color.a)) / 255;
			        bufp[1] = bufp[1] * (255-color.a) + (color.g * (color.a)) / 255;
		        	bufp[0] = bufp[0] * (255-color.a) + (color.b * (color.a)) / 255;
			}
		}
	}

	void drawImage(Point where, Image i, Color c = RGBA(255,255,255,255)){
		if(useGL && s.bpp == 124){
			int x = where.x;
			int y = where.y;
			int w = i.width;
			int h = i.height;

//			glColor4f(.5,.5,.5,1);
			glColor4f(c.r / 255.0, c.g / 255.0, c.b / 255.0, c.a / 255.0);
			glBindTexture(GL_TEXTURE_2D, i.toGL);
			glBegin(GL_QUADS); 
			glTexCoord2f(0, 0); glVertex2i(x, y);
			glTexCoord2f(i.texWidth, 0); glVertex2i(x+w, y); 
			glTexCoord2f(i.texWidth, i.texHeight); glVertex2i(x+w, y+h); 
			glTexCoord2f(0, i.texHeight); glVertex2i(x, y+h); 
			glEnd();

			glBindTexture(GL_TEXTURE_2D, 0); // unbind the texture... I guess
				// I don't actually understand why that is needed
				// but without it, everything drawn after it is wrong (too light or dark)
			return;
	}

		SDL_Rect r;
		r.x = where.x;
		r.y = where.y;
		r.w = i.width;
		r.h = i.height;
		if(locked)
			SDL_UnlockSurface(s.surface);

		if(SDL_BlitSurface(i.surface, null, s.surface, &r) == -1)
			throw new Error;

		if ( SDL_MUSTLOCK(s.surface) ) {
			if ( SDL_LockSurface(s.surface) < 0 ) {
				throw new Error;
			}
			locked = true;
		}

	}

	void drawText(Point where, char[] text, Color foreground = RGB(255,255,255), Color background = RGB(0,0,0), int font = 0){
		if(useGL && s.bpp == 124){
			Image i = FontEngine.instance.renderText(text, RGB(255,255,255), RGB(0,0,0), font);
			drawImage(where, i, foreground);
		}else{
			Image i = FontEngine.instance.renderText(text, foreground, background, font);
			drawImage(where, i);
		}
	}

	void drawTextCenteredHoriz(int top, char[] text, Color foreground, Color background, int font = 0){
		Point where;
		where.y = top;
		int w, h;
		TTF_SizeText(FontEngine.instance.fonts[font], std.string.toStringz(text), &w, &h);
		where.x = (s.width - w) / 2;
		drawText(where, text, foreground, background, font);
	}

	void line(Point start, Point end, Color color){
		if(useGL && s.bpp == 124){
			glColor4f(cast(float)(color.r)/255.0, cast(float)(color.g)/255.0, cast(float)(color.b)/255.0, cast(float)(color.a) / 255.0);
			glBegin(GL_LINES);
			glVertex2i(start.x, start.y);
			glVertex2i(end.x, end.y);
			glEnd();
		}
	}

	void hline(Point start, int width, Color color){
	if(useGL && s.bpp == 124){
		line(start, XY(start.x + width, start.y), color);
		return;
	}
		Point point = start;
		for(int a = 0; a < width; a++){
			putpixel(point, color);
			point.x++;
		}
	}

	void vline(Point start, int height, Color color){
	if(useGL && s.bpp == 124){
		line(start, XY(start.x, start.y + height), color);
		return;
	}

		Point point = start;
		for(int a = 0; a < height; a++){
			putpixel(point, color);
			point.y++;
		}
	}

	void circle(Point center, int radius, Color color){

	}

	void arc(Point center, int radius, float start, float end, Color color){
//		for(float a = start; a <= end; a+= (3.14159265358 / 50.0))
//			putpixel((int)(cos(a) * (float)radius + center.x()),(int)( sin(a) * (float) radius + center.y()), color);
	}

	void box(Point upperLeft, Point lowerRight, Color color){
		if(useGL && s.bpp == 124){
			int x1 = upperLeft.x;
			int y1 = upperLeft.y;
			int x2 = lowerRight.x;
			int y2 = lowerRight.y;
			glBegin(GL_QUADS);
			//glColor3b(color.r, color.g, color.b);
			glColor4f(cast(float)color.r/255.0, cast(float)color.g/255.0, cast(float)color.b/255.0, cast(float)color.a / 255.0);
			//glColor4f(1,1,1,1);
			glVertex2i(x1, y1);
			glVertex2i(x2, y1);
			glVertex2i(x2, y2);
			glVertex2i(x1, y2);
			glEnd();
			return;
		}
		SDL_Rect r;
		r.x = upperLeft.x;
		r.y = upperLeft.y;
		r.w = lowerRight.x - upperLeft.x;
		r.h = lowerRight.y - upperLeft.y;
		if(s.bpp == 32)
			SDL_FillRect(s.surface, &r, color.a << 24 | color.r << 16 | color.g << 8 | color.b);
		else
			SDL_FillRect(s.surface, &r, color.r << 16 | color.g << 8 | color.b);
	}

	void clear(){
		if(useGL && s.bpp == 124){
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT);
			return;
		}
		box(XY(0,0), XY(s.width, s.height), RGB(0,0,0));
	}

	void fill(Color color){
		box(XY(0,0), XY(s.width, s.height), color);
	}

	void blend(Color color){
		if(useGL && s.bpp == 124){
			box(XY(0,0), XY(s.width, s.height), color);
			return;
		}

		ubyte *bufp;

		bufp = cast(ubyte*)s.surface.pixels;
		for(int y = 0; y < s.height; y++)
			for(int x = 0; x < s.width; x++){
			
				bufp[2] = (bufp[2] * (255-color.a) + color.r * color.a) / 255;
			        bufp[1] = (bufp[1] * (255-color.a) + color.g * color.a) / 255;
		        	bufp[0] = (bufp[0] * (255-color.a) + color.b * color.a) / 255;

				bufp += (s.bpp == 24 ? 3 : 4);
			}
	}

  private:
  	Drawable s;
	bool locked;
}







int SDL_BlitSurface
			(SDL_Surface *src, SDL_Rect *srcrect,
			 SDL_Surface *dst, SDL_Rect *dstrect)
{
	return SDL_UpperBlit(src, srcrect, dst, dstrect);
}

bit SDL_MUSTLOCK(SDL_Surface *surface)
{
	return surface.offset || ((surface.flags &
		(SDL_HWSURFACE | SDL_ASYNCBLIT | SDL_RLEACCEL)) != 0);
}

/* Quick utility function for texture creation */
int power_of_two(int input)
{
    int value = 1;

    while ( value < input ) {
        value <<= 1;
    }
    return value;
}

uint SDL_GL_LoadTexture(SDL_Surface *surface, float *texcoord)
{
    uint texture;
    int w, h;
    SDL_Surface *image;
    SDL_Rect area;
    uint saved_flags;
    ubyte  saved_alpha;

    /* Use the surface width and height expanded to powers of 2 */
    w = power_of_two(surface.w);
    h = power_of_two(surface.h);
    texcoord[0] = 0.0f;         /* Min X */
    texcoord[1] = 0.0f;         /* Min Y */
    texcoord[2] = cast(float)surface.w / cast(float)w;  /* Max X */
    texcoord[3] = cast(float)surface.h / cast(float)h;  /* Max Y */

    image = SDL_CreateRGBSurface(
            SDL_SWSURFACE,
            w, h,
            32,
            0x000000FF,
            0x0000FF00,
            0x00FF0000,
            0xFF000000
               );
    if ( image == null) {
        throw new Error;
    }


    /* Save the alpha blending attributes */
    saved_flags = surface.flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
    saved_alpha = surface.format.alpha;
    if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
        SDL_SetAlpha(surface, 0, 0);
    }

    /* Copy the surface into the GL texture image */
    area.x = 0;
    area.y = 0;
    area.w = surface.w;
    area.h = surface.h;
    SDL_BlitSurface(surface, &area, image, &area);

    /* Restore the alpha blending attributes */
    if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
        SDL_SetAlpha(surface, saved_flags, saved_alpha);
    }

    /* Create an OpenGL texture for the image */
    glGenTextures(1, &texture);

    glBindTexture(GL_TEXTURE_2D, texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    glTexImage2D(GL_TEXTURE_2D,
             0,
             GL_RGBA,
             w, h,
             0,
             GL_RGBA,
             GL_UNSIGNED_BYTE,
             image.pixels);
    SDL_FreeSurface(image); /* No longer needed */



    return texture;
}
